#!/usr/bin/python
import colorsys
import cv2
import itertools
import numpy as np
import random
import tensorflow as tf
import time

from core.config import cfg
from logger import get_logger


logger = get_logger(__name__)

def decode_label(out):
    CHAR_VECTOR = "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
    letters = [letter for letter in CHAR_VECTOR]
    out_best = np.argmax(out[0, 2:], axis=1)
    confidence = np.take_along_axis(out[0, 2:], out_best[:, None], axis=-1)
    out_best = np.column_stack((out_best, confidence))

    new_out_best = []
    for key, group in itertools.groupby(out_best, lambda x: x[0]):
        max_confidence = 0
        for conf in group:
            if conf[1] > max_confidence:
                max_confidence = conf[1]
            else:
                continue
        new_out_best.append([int(key), max_confidence])

    outstr = []
    for index, values in enumerate(new_out_best):
        if values[0] < len(letters):
            outstr.append([letters[values[0]], values[1]])

    return outstr


def read_plate(img, coords, model):
    final_pred = ""
    confidence = []
    try:
        start_trp = time.perf_counter()
        xmin, ymin, xmax, ymax = coords
        box = img[int(ymin) - 5: int(ymax) + 5, int(xmin) - 5: int(xmax) + 5]
        time_box = time.perf_counter()

        img_pred = cv2.cvtColor(box, cv2.COLOR_RGB2GRAY)
        time_img_gray = time.perf_counter()
        img_pred = img_pred.astype(np.float32)
        time_img_f32 = time.perf_counter()
        img_pred = cv2.resize(img_pred, (128, 64))
        time_img_resize = time.perf_counter()

        img_pred = (img_pred / 255.0) * 2.0 - 1.0
        time_img_bin = time.perf_counter()
        img_pred = img_pred.T
        time_img_T = time.perf_counter()
        img_pred = np.expand_dims(img_pred, axis=-1)
        time_img_np_ex_1 = time.perf_counter()
        img_pred = np.expand_dims(img_pred, axis=0)
        time_img_np_ex_2 = time.perf_counter()
        net_out_value = model.predict(img_pred)
        time_img_model = time.perf_counter()
        pred_texts = decode_label(net_out_value)
        time_img_label = time.perf_counter()

        for char in pred_texts:
            final_pred += char[0]
            confidence.append(np.round(char[1], 2))
        time_img_char = time.perf_counter()
        time_img_total_trp = time.perf_counter()

        total_time_box = round((time_box - start_trp) * 1000, 3)
        total_img_gray = round((time_img_gray - time_box) * 1000, 3)
        total_img_f32 = round((time_img_f32 - time_img_gray) * 1000, 3)
        total_img_resize = round((time_img_resize - time_img_f32) * 1000, 3)
        total_img_bin = round((time_img_bin - time_img_resize) * 1000, 3)
        total_img_T = round((time_img_T - time_img_bin) * 1000, 3)
        total_img_np_ex_1 = round((time_img_np_ex_1 - time_img_T) * 1000, 3)
        total_img_np_ex_2 = round((time_img_np_ex_2 - time_img_np_ex_1) * 1000, 3)
        total_img_model = round((time_img_model - time_img_np_ex_2) * 1000, 3)
        total_img_label = round((time_img_label - time_img_model) * 1000, 3)
        total_img_char = round((time_img_char - time_img_label) * 1000, 3)
        total_img_total_trp = round((time_img_total_trp - start_trp) * 1000, 3)

        logger.debug(
            f'{f"Tempo Box: {total_time_box}| "}'
            f'{f"Tempo Gray: {total_img_gray}| "}'
            f'{f"Tempo Float32: {total_img_f32}| "}'
            f'{f"Resize: {total_img_resize}| "}'
            f'{f"Binarização: {total_img_bin}| "}'
            f'{f"T: {total_img_T}| "}'
            f'{f"Expand Dims1: {total_img_np_ex_1}| "}'
            f'{f"Expand Dims2: {total_img_np_ex_2}| "}'
            f'{f"Model: {total_img_model}| "}'
            f'{f"Label: {total_img_label}| "}'
            f'{f"Char: {total_img_char}| "}'
            f'{f"Total TRP: {total_img_total_trp}"}')

        return final_pred, confidence
    except:
        return final_pred, confidence


def read_class_names(class_file_name):
    names = {}
    with open(class_file_name, "r") as data:
        for ID, name in enumerate(data):
            names[ID] = name.strip("\n")
    return names


def load_config():
    STRIDES = np.array(cfg.YOLO.STRIDES)
    ANCHORS = get_anchors(cfg.YOLO.ANCHORS)
    XYSCALE = cfg.YOLO.XYSCALE
    NUM_CLASS = len(read_class_names(cfg.YOLO.CLASSES))
    return STRIDES, ANCHORS, NUM_CLASS, XYSCALE


def get_anchors(anchors_path):
    anchors = np.array(anchors_path)
    return anchors.reshape(3, 3, 2)


def image_preprocess(image, target_size):
    ih, iw = target_size
    h, w, _ = image.shape

    scale = min(iw / w, ih / h)
    nw, nh = int(scale * w), int(scale * h)
    image_resized = cv2.resize(image, (nw, nh))

    image_paded = np.full(shape=[ih, iw, 3], fill_value=128.0)
    dw, dh = (iw - nw) // 2, (ih - nh) // 2
    image_paded[dh: nh + dh, dw: nw + dw, :] = image_resized
    image_paded = image_paded / 255.0

    image_for_model = image_paded[np.newaxis, ...].astype(np.float32)
    batch_data = tf.constant(image_for_model)

    return batch_data, image.copy()


def format_boxes(bboxes, image_height, image_width):
    for box in bboxes:
        ymin = int(box[0] * image_height)
        xmin = int(box[1] * image_width)
        ymax = int(box[2] * image_height)
        xmax = int(box[3] * image_width)
        box[0], box[1], box[2], box[3] = xmin, ymin, xmax, ymax
    return bboxes


def draw_bbox(image, bboxes, ocr_model, plate=False):
    classes = read_class_names(cfg.YOLO.CLASSES)
    num_classes = len(classes)
    image_h, image_w, _ = image.shape
    hsv_tuples = [(1.0 * x / num_classes, 1.0, 1.0) for x in range(num_classes)]
    colors = list(map(lambda x: colorsys.hsv_to_rgb(*x), hsv_tuples))
    colors = list(
        map(lambda x: (int(x[0] * 255), int(x[1] * 255), int(x[2] * 255)), colors)
    )

    random.seed(0)
    random.shuffle(colors)
    random.seed(None)

    plate_number = ""
    confidence = []
    out_boxes, out_scores, out_classes, num_boxes = bboxes

    if num_boxes >= 2:
        plate_th = 0.8
    else:
        plate_th = 0.3

    for i in range(num_boxes):
        if int(out_classes[i]) < 0 or int(out_classes[i]) > num_classes:
            continue
        if out_scores[i] < plate_th:
            continue
        coor = out_boxes[i]
        # fontScale = 0.5
        score = round((out_scores[i]) * 100, 0)
        logger.debug(
            f'{f"Score plate model: {score}"}'
        )
        # class_ind = int(out_classes[i])
        # class_name = classes[class_ind]

        if plate:
            plate_number, confidence = read_plate(image, coor, ocr_model)

    return image, plate_number, confidence


def bboxes_iou(boxes1, boxes2):
    """
    Calcula o IOU entre dois arrays de bounding boxes
    """
    boxes1 = np.array(boxes1)
    boxes2 = np.array(boxes2)

    boxes1_area = (boxes1[..., 2] - boxes1[..., 0]) * (boxes1[..., 3] - boxes1[..., 1])
    boxes2_area = (boxes2[..., 2] - boxes2[..., 0]) * (boxes2[..., 3] - boxes2[..., 1])

    left_up = np.maximum(boxes1[..., :2], boxes2[..., :2])
    right_down = np.minimum(boxes1[..., 2:], boxes2[..., 2:])

    inter_section = np.maximum(right_down - left_up, 0.0)
    inter_area = inter_section[..., 0] * inter_section[..., 1]
    union_area = boxes1_area + boxes2_area - inter_area
    ious = np.maximum(1.0 * inter_area / union_area, np.finfo(np.float32).eps)

    return ious


def get_pred_bbox_to_recognition(bboxes):
    deltax = 5
    deltay = 5
    boxes_list = []
    scores_list = []
    class_list = []
    for bbox in bboxes:
        boxes_list.append([bbox[0] - deltax, bbox[1] - deltay, bbox[2] + deltax, bbox[3] + deltay])
        scores_list.append(bbox[4])
        class_list.append(bbox[5])

    array_bbox = np.array(boxes_list)
    array_score = np.array(scores_list)
    array_class = np.array(class_list)
    n_detections = len(array_bbox)

    pred_bbox = [array_bbox, array_score, array_class, n_detections]

    return pred_bbox


def process_detection_prediction(pred_bbox, original_image, input_size):
    # Processa a predição para obter bboxes válidas

    iou_threshold = 0.2
    score_threshold = 0.6
    valid_scale = [0, np.inf]

    # Obter resultados
    pred_bbox = np.array(pred_bbox)
    pred_xywh = pred_bbox[:, 0:4]
    pred_conf = pred_bbox[:, 4]
    pred_prob = pred_bbox[:, 5:]

    # Conversão do formato
    pred_coor = np.concatenate([pred_xywh[:, :2] - pred_xywh[:, 2:] * 0.5,
                                pred_xywh[:, :2] + pred_xywh[:, 2:] * 0.5], axis=-1)

    # Ajuste das coordenadas
    org_h, org_w = original_image.shape[:2]
    resize_ratio = min(input_size / org_w, input_size / org_h)

    dw = (input_size - resize_ratio * org_w) / 2
    dh = (input_size - resize_ratio * org_h) / 2

    pred_coor[:, 0::2] = 1.0 * (pred_coor[:, 0::2] - dw) / resize_ratio
    pred_coor[:, 1::2] = 1.0 * (pred_coor[:, 1::2] - dh) / resize_ratio

    # Clipar bbox se necessário
    pred_coor = np.concatenate([np.maximum(pred_coor[:, :2], [0, 0]),
                                np.minimum(pred_coor[:, 2:], [org_w - 1, org_h - 1])], axis=-1)
    invalid_mask = np.logical_or((pred_coor[:, 0] > pred_coor[:, 2]), (pred_coor[:, 1] > pred_coor[:, 3]))
    pred_coor[invalid_mask] = 0

    bboxes_scale = np.sqrt(np.multiply.reduce(pred_coor[:, 2:4] - pred_coor[:, 0:2], axis=-1))
    scale_mask = np.logical_and((valid_scale[0] < bboxes_scale), (bboxes_scale < valid_scale[1]))

    # Obter pred bbox
    classes = np.argmax(pred_prob, axis=-1)
    scores = pred_conf * pred_prob[np.arange(len(pred_coor)), classes]
    score_mask = scores > score_threshold
    mask = np.logical_and(scale_mask, score_mask)
    coors, scores, classes = pred_coor[mask], scores[mask], classes[mask]

    bboxes = np.concatenate([coors, scores[:, np.newaxis], classes[:, np.newaxis]], axis=-1)

    classes_in_img = list(set(bboxes[:, 5]))
    best_bboxes = []

    for classe in classes_in_img:
        cls_mask = (bboxes[:, 5] == classe)
        cls_bboxes = bboxes[cls_mask]

        # Se houver bounding boxes:
        while len(cls_bboxes) > 0:
            # Selecionar as bbox com o maior score
            max_ind = np.argmax(cls_bboxes[:, 4])
            best_bbox = cls_bboxes[max_ind]
            best_bboxes.append(best_bbox)
            cls_bboxes = np.concatenate([cls_bboxes[: max_ind], cls_bboxes[max_ind + 1:]])

            # Calcular o IOU entre essa bbox e as demais bbox e
            # remover aquelas cujo IOU é maior que o threshold estabelecido.
            iou = bboxes_iou(best_bbox[np.newaxis, :4], cls_bboxes[:, :4])
            weight = np.ones((len(iou),), dtype=np.float32)

            iou_mask = iou > iou_threshold
            weight[iou_mask] = 0.0

            cls_bboxes[:, 4] = cls_bboxes[:, 4] * weight
            score_mask = cls_bboxes[:, 4] > 0.
            cls_bboxes = cls_bboxes[score_mask]

    return best_bboxes
